'SDCard YModem tranfer and utilities
'(C) 2008 Raymond Allen
' Free for non-commercial (except Prop-OLED Contest!) usage as long as this file and notice are not altered
' Not extensively tested!  Use at own risk!
'
CON  'Constants Section

  _clkmode = xtal1 + pll16x      'using 5 MHz crystal in pll16x mode
  _xinfreq = 5_000_000           'to achieve 80 Mhz clock frequency 

  SD_Basepin=0  'basepin for SD card driver

  Baudrate=115200    'baud rate for serial communications
  RXPIN=31        'Prop pins used for comms
  TXPIN=30

  'XMODEM chars from ymodem.txt
  SOH=$01
  STX=$02
  EOT=$04
  ACK=$06
  NAK=$15
  CAN=$18
  Cee=$43  'liberties here
  Zee=$1A  'or SUB  (DOS EOF?)

OBJ  'Objects Section

  ser : "FullDuplexSerialExtended"  'for serial comms with big buffer
  sdfat : "fsrw_Ray"  'Rockiki's OBEX SD card interface (modified to give filesizes)
  num : "numbers"  'to format numbers 

var
   byte tbuf[20]
   byte stack[100]
   long Mounted
   byte fbuf[30]
   byte sbuf[30]
   byte pdata[1028] 'packet data

PUB Main|cog,key,crc,i,z   


  num.init  'start numbers
  ser.start(RXPIN,TXPIN,0,BaudRate) 'open comm port


   'Establish serial link (wait for client to hit spacebar)
  key:=0
  repeat until key==32
    waitcnt(cnt+clkfreq)
    ser.str(string(27,"[2J Hit spacebar to begin.  "))',10,13))
    repeat 
      key:=ser.rxcheck
    while key>0 and key<>32
    
  'Show welcome logo
  ser.str(string(27,"[2J Welcome to the YModem file transfer utility!",10,13))

  'use a new cog to try to mount SD card (because it can hang if it fails)
  Mounted:=-1
  repeat until Mounted==0
    cog:=cognew(Mount,@stack)
    waitcnt(cnt+clkfreq/2)  'give it a moment
    cogstop(cog)  'force it to terminate
     
    if Mounted<>0
      Waitkey(32,string("SD Card Mount Failed:  Press spacebar to retry",10,13))
      'key:=0
      'repeat until key==32
      '  waitcnt(cnt+clkfreq)
      '  ser.str(string("Mount Failed.  Press spacebar to retry.  ",10,13))
      '  repeat 
      '    key:=ser.rxcheck
      '  while key>0 and key<>32      

  waitcnt(cnt+clkfreq/2)  'let user soak it up for a second

  Repeat
    'Show welcome logo again
    ser.str(string(27,"[2J      YModem Utility Menu",10,13))
    ser.str(string(       "=============================", 13,10,13))
    ser.str(string(       "D = Directory of SD card", 13,10))
    ser.str(string(       "S = Send files to SD card", 13,10))
    ser.str(string(       "R = Receive a file from SD card", 13,10))
    ser.str(string(       "X = Erase file on SD card", 13,10)) 
    ser.str(string(       "Q = Quit", 13,10))   
    key:=0
    repeat
      key:=ser.rxcheck
    until key>0
    case key
      "D","d": 'directory
        dir
      "Q","q": 'quit
        ser.str(string(27,"[2J Quitted."))
        sdfat.stop
        ser.stop
        repeat
      "R","r":  'receive file
        Receive 
      "S","s":  'send file
        Send
      "X","x":  'erase file
        Erase
         
     
    Waitkey(32,string(13,10,"Press spacebar to continue",13,10))

PUB erase|key,i
  ser.str(string(13,10,"Erase a file from SD card."))
  ser.str(string(13,10,">Enter filename in 8.3 format (press ESC to abort):"))
  key:=0
  i:=0
  
  repeat until key==13
    key:=ser.rx
    if (key==27)
      ser.str(string(13,10,"Aborted by keystroke."))
      return 'abort
    ser.tx(key)
    if key<>13
      fbuf[i++]:=key
    else
      fbuf[i]:=0  'append end of string mark

  if strsize(@fbuf)>12
      ser.str(string(13,10,"Filename too long!  Use 8.3 format."))
      return 'fail

  ser.str(string(13,10,"Are you sure? (Y/N)"))
  i:=ser.rx
  if i<>"Y" and i<>"y"
    ser.str(string(13,10,"Aborted by keystroke."))
    return

  i:=sdfat.popen(@fbuf,"d")
  if i<0
    ser.str(string(13,10,"Erase Failed."))
  else
    ser.str(string(13,10,"Erase Complete."))

  return  

      

PUB send|key,i,packet,crc,j,k,timer,bytes,done
  'get file from computer and save to SD
  'Ymodem protocol: http://timeline.textfiles.com/1988/10/14/1/FILES/ymodem.txt

  ser.str(string(13,10,"Send file via ymodem now..."))
  repeat  'batch reception loop
    'wait for first packet
    packet:=0
    repeat until packet==1
      key:=0
      'repeat until key==SOH
        waitcnt(cnt+clkfreq*10)
        ser.tx(Cee)
        'waitcnt(cnt+clkfreq*1)
        timer:=cnt
        repeat until i==SOH
          i:=ser.rxcheck

          if i>0
            key:=i
          if (cnt-timer)>clkfreq*5
            ser.tx(EOT)
            ser.str(string("No response from host... Aborting."))
            return
      'analyze first packet
      if (ser.rx==0)
        if (ser.rx==$FF)

          crc:=0
          j:=0
          i:=-1
          done:=0
          'filename
          repeat until i==0
            i:=ser.rx
            if i==0 and j==0
              'End of session
              done:=1
            fbuf[j++]:=i
            crc:=UpdateCRC(i,crc)
            
          i:=-1
          k:=j
          j:=0

          repeat until i==0 or i==32
            i:=ser.rx
            if i<>32
              sbuf[j++]:=i
            else
              sbuf[j++]:=0
            crc:=UpdateCRC(i,crc)
          k+=j

          bytes:=num.fromstr(@sbuf,num#dec)
     

          
          repeat j from k+1 to 128
            i:=ser.rx
            crc:=UpdateCRC(i,crc)
        else
          ser.tx(NAK)
      else
        ser.tx(NAK)
 
      i:=ser.rx

      j:=ser.rx

      if i<>(crc>>8) or j<>(crc&$FF)
        ser.tx(NAK)

        
      else
        packet:=1
     

    ser.tx(ACK)
     
    if done==1
      ser.str(string("Batch file reception complete."))
      return 
     
     
    ser.tx(Cee)
     
    'open output file
    if sdfat.popen(@fbuf,"w")<0

      ser.tx(EOT)
      ser.str(string("Can't open file... Aborting."))
      return
      
    'receive packets
    k:=0
    repeat until k==EOT
      k:=ser.rx
      if k==SOH
        '128 byte packets
        packet:=ser.rx
        i:=ser.rx
        if (255-i)<>packet

          repeat
        else
          'text.str(string("Good packet header"))
        crc:=0
        repeat j from 0 to 127

          i:=ser.rx
          crc:=UpdateCRC(i,crc)
          pdata[j]:=i
        i:=ser.rx
        j:=ser.rx
        if i<>(crc>>8) or j<>(crc&$FF)
          ser.tx(NAK)

        else
          'write data
          if bytes>128
            sdfat.pwrite(@pdata,128)
            bytes-=128
          else
            sdfat.pwrite(@pdata,bytes)
            bytes:=0
          'get more
          ser.tx(ACK)
      else
        if k==STX
          '1024 byte packets
          repeat j from 0 to 1027
            pdata[j]:=ser.rx
           
          packet:=pdata[0]
          i:=pdata[1]
          if (255-i)<>packet

            repeat


          crc:=0
          repeat j from 2 to 1025
            i:=pdata[j]
            crc:=UpdateCRC(i,crc)
          i:=pdata[1026]
          j:=pdata[1027]
          if i<>(crc>>8) or j<>(crc&$FF)
            ser.tx(NAK)
          else
            'write data
            if bytes>1024
              sdfat.pwrite(@pdata+2,1024)
              bytes-=1024
            else
              sdfat.pwrite(@pdata+2,bytes)
              bytes:=0
            'get more
            ser.tx(ACK)

          

    ser.tx(ACK)
    sdfat.pclose 
  
      
                
  
    
    
  

  

PUB receive|key,i,size,seconds,timer,packet,d,crc,p,j,k,deot
  'Ymodem protocol: http://timeline.textfiles.com/1988/10/14/1/FILES/ymodem.txt
  '                 http://en.wikipedia.org/wiki/XMODEM
  
  ser.str(string(13,10,"Receive a file from SD card."))
  ser.str(string(13,10,">Enter filename in 8.3 format (press ESC to abort):"))
  key:=0
  i:=0
  
  repeat until key==13
    key:=ser.rx
    if (key==27)
      ser.str(string(13,10,"Aborted by keystroke."))
      return 'abort
    ser.tx(key)
    if key<>13
      fbuf[i++]:=key
    else
      fbuf[i]:=0  'append end of string mark

  if strsize(@fbuf)>12
      ser.str(string(13,10,"Filename too long!  Use 8.3 format."))
      return 'fail       


  'Transmit file to host
  if sdfat.popen(@fbuf,"r")<>0
    ser.str(string(13,10,"File not found!"))
    return 'fail
  size:=sdfat.fsize
  ser.str(string(13,10,"Transmitting file.  Instruct terminal to receive YMODEM file now."))

  
  timer:=cnt
  key:=0
  'Wait for NAK from host:  Signal to start sending
  repeat until key==Cee
    i:=ser.rxcheck
    if i>0
      key:=i
    i:=cnt
    if ((cnt-timer)/clkfreq)>10           '10 second timeout
      ser.str(string(13,10,"Failure.  No response from host.."))
      return 'fail  

  
  'start sending packets
  packet:=0
  repeat
    timer:=cnt
    crc:=0
    deot:=false  
    'send header
    ser.tx(SOH)
    ser.tx(packet)
    ser.tx(!packet)

    'construct packet
    if packet==0
      'Send filename and length
      i:=strsize(@fbuf)
      bytemove(@pdata,@fbuf,i+1)
      p:=num.tostr(size,num#DEC)
      j:=strsize(p)
      bytemove(@pdata+i+1,p,j+1)
      repeat k from i+1+j+1 to 128
        pdata[k]:=0
    else  
      'data packet
      repeat i from 0 to 127
        d:=sdfat.pgetc
        if d==-1
          'end of file
          repeat i from i to 127
            d:=EOT'Zee
            pdata[i]:=d
            deot:=true
          quit          
        else
          if d<0 
            d:=EOT
            ser.tx(d)
            return 'fail      

        pdata[i]:=d

    'send packet
    repeat i from 0 to 127
      ser.tx(pdata[i])

      crc:=UpdateCRC(pdata[i],crc)
       

    ser.tx((crc>>8)&$FF)
    ser.tx(crc&$FF)
    

    repeat
      i:=ser.rxcheck
      if i==ACK
        quit
      if i==NAK
        'retransmit packet
        repeat i from 0 to 127
          ser.tx(pdata[i])
          ser.tx((crc>>8)&$FF)
          ser.tx(crc&$FF)
      if ((cnt-timer)/clkfreq)>10
        ser.tx(EOT)
        ser.str(string(13,10,"Timeout failure."))
        return 'fail
    if (deot==true)'Zee) 'done
      quit
    else
      packet++
      timer:=cnt
    

  'send EOT wait for ACK and send end of batch packet
  ser.tx(EOT)
  repeat
    i:=ser.rxcheck
    if i==ACK
      'now, wait for "C"
      timer:=cnt
      key:=0
      repeat until key==Cee
        i:=ser.rxcheck
        if i>0
          key:=i
        i:=cnt
        if ((cnt-timer)/clkfreq)>10           '10 second timeout
          ser.str(string(13,10,"Failure.  No response from host.."))
          return 'fail
      'now, transmit null packet and wait for ACK
      key:=0
      repeat until key==ACK
        ser.tx(SOH)
        ser.tx(0)
        ser.tx($FF)
        crc:=0
        repeat 128
          ser.tx(0)
          crc:=UpdateCRC(0,crc)
        ser.tx((crc>>8)&$FF)
        ser.tx(crc&$FF)
        'wait for ack
        key:=ser.rx
        if key<>ACK and key<>NAK
          return 'fail  
      quit  'all done!
    if i>0
      ser.tx(EOT)

  ser.str(string(13,10,"YModem receive from SD complete.",13,10))

    
      
    
PRI UpdateCRC(data,crc):newcrc|i,icrc
 'look here:http://web.mit.edu/6.115/www/miscfiles/amulet/amulet-help/xmodem.htm

  crc:=crc^(data<<8)
  repeat i from 0 to 7
    if crc&$8000
      crc:=((crc<<1)&$FFFF)^$1021
    else
      crc:=(crc<<=1)&$FFFF

  return crc&$FFFF


          
      
      


PUB Mount|key

  'Try to mount SD card

  Mounted:=sdfat.mount(SD_Basepin)
  case Mounted
    -20: ' not a fat16 volume
      ser.str(string("Error:  Not a FAT16 volume.",10,13))
    -21: ' bad bytes per sector
      ser.str(string("Error:  Bad bytes per sector.",10,13))
    -22: ' bad sectors per cluster  
      ser.str(string("Error:  Bad sectors per cluster.",10,13))
    -23: ' bad bytes per sector
      ser.str(string("Error:  Missing second FAT.",10,13))
    -24: ' bad FAT signature
      ser.str(string("Error:  Bad FAT signature.",10,13))
    -25: ' bad FAT signature
      ser.str(string("Error:  Too many clusters.",10,13))
    0:  'OK
      ser.str(string("SD card mounted successfully.",10,13))
  




pub dir|lines,key,total,size,i
  lines:=0   
  ser.str(string(27,"[2J Directory of SD Card: ", 13,10))
  ser.str(string("===================================", 13,10))
  ser.str(string("Filename        Size in Bytes", 13,10))
  ser.str(string("------------    -------------", 13,10))

  
  sdfat.opendir
  repeat while 0 == sdfat.nextfile(@tbuf)
    repeat 12-strsize(@tbuf)
     ser.tx(" ")
    ser.str(@tbuf)
    ser.str(string("   "))
    size:=sdfat.fsize
    total+=size
    ser.str(num.tostr(size,num#DSDEC14))
    ser.tx(13)
    ser.tx(10)
    if ++lines>20
      Waitkey(32,string("Press spacebar to continue",13,10))
      lines:=0

  'Show total bytes  
  ser.str(string(13,10,10,"Total: "))
  ser.str(num.tostr(total,num#DSDEC14))   
  ser.str(string(13,10,10,"bytes"))
      

Pub Waitkey(k,str)|key  'show string and wait for key=k
  ser.str(str)
  key:=0
  repeat until key==k
    repeat 
      key:=ser.rxcheck
    while key>0 and key<>k   


Dat